#define EXT_USESCIBENCH

#define MEASALL
#include <sstream>
#include <string>
#include <tmpi.h>
#include <debug/debugger.h>
#include <random>
#include <algorithm>

#include <unistd.h>

#include <cassert>

struct elem {
	int start;
	int num;
	int idx;
	int r_idx;
};

// lets do a larger test. N messages, arbitrary size around rdv-eager limit
// lets use 1024 messages with size around 256
int main(int argc, char** argv) {

	int rank;
	int size;

	constexpr int N = 182;
	constexpr int MES_N = 256;
	const int length = MES_N*N;
	int* message = new int[length];
	for (int i = 0; i < length; ++i)
		message[i] = i;

	assert(sizeof(int) == 4);

	TMPI_Init(&argc, &argv);

	TMPI_Comm_size(TMPI_COMM_WORLD, &size);
	TMPI_Comm_rank(TMPI_COMM_WORLD, &rank);

	TMPI_Request tr[N]; // we just do three requests
	TMPI_Status st[N];

	std::mt19937 mt(12);
	// generate 16*N random numbers
	std::vector<int> random_numbers;
	for (int i = 0; i < length; ++i)
		random_numbers.push_back(i);
	std::shuffle(random_numbers.begin(), random_numbers.end(), mt);

	std::sort(random_numbers.begin(), random_numbers.begin() + N);
	for (int i = 0; i < N; ++i)
		TMPI_LOG(random_numbers[i], i);

	// and take the first N-1 values
	int* starters = new int[N];
	starters[0] = 0;
	for (int i = 0; i < N-1; ++i)
		starters[i+1] = random_numbers[i];

	// from which we calculate the sizes
	int* sizes = new int[N];
	for (int i = 0; i < N-1; ++i)
		sizes[i] = starters[i+1] - starters[i];
	sizes[N-1] = length - starters[N-1];

	// now we create a mapping. the nth send is the nth entry
	std::vector<int> mapping(N);
	for (int i = 0; i < N; ++i)
		mapping[i] = i;
	std::random_device rd;
	std::mt19937 mt_random(rd());
	std::shuffle(mapping.begin(), mapping.end(), mt_random);

	std::vector<elem> elems(N);
	for (int i = 0; i < N; ++i) {
		elems[i].start = starters[i];
		elems[i].num = sizes[i];
		elems[i].r_idx = mapping[i];
		elems[i].idx = i;
	}
	delete[] starters;
	delete[] sizes;

	std::sort(elems.begin(), elems.end(), [] (elem& a, elem& b) { return a.r_idx < b.r_idx; });

	// just send to the destination
	auto do_sends_to = [&](int dest) {
		for (int i = 0; i < N; ++i) {

			TMPI_Isend(message + elems[i].start, elems[i].num, TMPI_INT, dest, elems[i].idx, TMPI_COMM_WORLD, 
					&tr[i]);
		}
		
		TMPI_Waitall(N, tr, st);
	};
	// recv from source
	auto do_recvs_from = [&](int source) {
		for (int i = 0; i < N; ++i) {
			TMPI_Irecv(message + elems[i].start, elems[i].num, TMPI_INT, source, elems[i].idx, TMPI_COMM_WORLD,
					&tr[i]);
		}
		
		TMPI_Waitall(N, tr, st);
	};
	auto do_sends = [&]() { 
		do_sends_to((rank + 1) % size);
	};
	auto do_rsends = [&]() { 
		do_sends_to((rank + size - 1) % size);
	};
	auto do_recvs = [&]() { 
		do_recvs_from((rank + size - 1) % size);
	};
	auto do_rrecvs = [&]() { 
		do_recvs_from((rank + 1) % size);
	};

	constexpr unsigned numreps = 1000;
	for (unsigned reps = 0; reps < numreps; ++reps) {
		if (rank == 0) {
			std::cout << reps << std::endl;
			do_sends();
			std::cout << reps << std::endl;
			do_recvs();
			std::cout << reps << std::endl;

			do_rsends();
			do_rrecvs();

		} else {
			do_recvs();
			for (int i = 0; i < length; ++i)
				message[i] += 1;
			do_sends();

			do_rrecvs();
			do_rsends();
		}
	}

	for (unsigned i = 0; i < length; ++i) {
		using namespace tinympi;
		TMPI_ASSERT_FATAL(message[i] == (int)(i + (size-1)*numreps), "wrong output! Expected "
				<< i + (size-1)*numreps << ", but got " << message[i],
				__FILE__, __LINE__);
	}

	TMPI_Finalize();

	delete[] message;

}
