// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#include "global.h"
#include "pt2pt.h"
#include "debug/debugger.h"

namespace tinympi {
namespace collectives {
namespace impl {
	using namespace tinympi::impl;

/////////////////////////////////////
// BCAST
namespace bcast {
inline int bcast_binomial(void* buffer, int length, int root, const Communicator* comm) {
	int res;

	rank_type my_rank = comm->get_rank();
	rank_type size = comm->get_size();

	// no we have original order 0    1 2 3 4 ... root-1 root root+1 ...   n-1
	// and relative order      n-root    ...     n-1   0     1   ...  n-root-1
	rank_type rel_rank = my_rank + (my_rank < root ? size : 0) - root;

	// we have exactly one receive (except for the root). this is given by the least significant bit,
	// i.e. 0b010000101000
	// -->  0b010000100000
	//                ^
	rank_type mask = 1;
	while (mask < size) {
		if ((mask & rel_rank) > 0) {
			rank_type src = my_rank - mask;
			if (src < 0)
				src += size;

			TMPI_LOG_LEVEL(-1, "Collectives", "recv from " << src);
			res = pt2pt::recv(buffer, length, src, TMPI_BCAST_TAG, comm->get_id(), comm,
					TMPI_STATUS_IGNORE);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "recv() failed", __FILE__, __LINE__);
			break;
		}
		mask <<= 1;
	}

	// then post the corresponding sends
	mask >>= 1; // we start with the most significant bit
	TMPI_LOG_LEVEL(-1, "Collectives", "start send mask with " << mask);
	while (mask > 0) {
		if (rel_rank + mask < size) {
			rank_type dest = my_rank + mask;
			if (dest >= size)
				dest -= size;

			TMPI_LOG_LEVEL(-1, "Collectives", "send to " << dest);
			res = pt2pt::send(buffer, length, dest, TMPI_BCAST_TAG, comm->get_id(), comm);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "send() failed", __FILE__, __LINE__);
		}
		mask >>= 1;
	}

	return TMPI_SUCCESS;
}
} // END bcast

}
}
}
