// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _BASEREQUEST_H_635106f3226c46d3d114158785752466
#define _BASEREQUEST_H_635106f3226c46d3d114158785752466

#include <malloc.h>
#include <cstddef>

#include <boost/pool/pool.hpp>

#include "tasks.h"
#include "communicator.h"
#include "global.h"
#include "debug/debugger.h"

namespace tinympi {
namespace impl {

// simplify the way we can convert between different request types.
#define VERIFY_COMPATIBILITY(TYPE1, TYPE2, ELEMENT) \
	static_assert(offsetof(TYPE1, ELEMENT) == offsetof(TYPE2, ELEMENT), \
			#TYPE1 " and " #TYPE2 " are not compatible! (Element: " #ELEMENT ")");

#define REQUESTS_CONVERTIBLE(type1, type2) \
	inline type1* TO_ ## type1(type2* in) { \
		return reinterpret_cast<type1*>(in); \
	} \
	inline type2* TO_ ## type2(type1* in) { \
		return reinterpret_cast<type2*>(in); \
	}


// this are the methods that should be implemented for each request
enum class dispatch_methods {
	posted_request_for_unexpected, // only needed on provider level
	fill_status,
	delete_request,
	try_cancel_request
};
inline std::string to_string(dispatch_methods method) {
	switch (method) {
	case dispatch_methods::posted_request_for_unexpected:
		return "dispatch_methods::posted_request_for_unexpected";
	case dispatch_methods::fill_status:
		return "dispatch_methods::fill_status";
	case dispatch_methods::delete_request:
		return "dispatch_methods::delete_request";
	case dispatch_methods::try_cancel_request:
		return "dispatch_methods::try_cancel_request";
	default:
		return "unknown dispatch type";
	}
}

enum class completion_status_type : uint16_t {
	REQUEST_INPROGRESS,
	REQUEST_COMPLETED,
	REQUEST_CANCELLED
};
inline std::string to_string(completion_status_type cs) {
	switch(cs) {
	case completion_status_type::REQUEST_INPROGRESS:
		return "completion_status_type::REQUEST_INPROGRESS";
	case completion_status_type::REQUEST_COMPLETED:
		return "completion_status_type::REQUEST_COMPLETED";
	case completion_status_type::REQUEST_CANCELLED:
		return "completion_status_type::REQUEST_CANCELLED";
	default:
		return "";
	}
}
union atomic_state_type {
	struct {
		completion_status_type completion_status;
		uint16_t waiting;
	} data;
	uint32_t asInt;
};

enum class provider_types : uint16_t {
	node2node,
	thread2thread,
	task2task,
	self2self,
};

#define BASE_REQUEST_MEMBERS \
	std::atomic<atomic_state_type> atomic_state; /* 4 */ \
	provider_types provider_type;                /* 2 */ \
	context_id_t context_id;                     /* 2 */ \
	global_id_t peer_rank;                       /* 8 */ \
	tag_t tag;                                   /* 4 */ \
	buflen_t buflen;                             /* 4 */ \
	void* buf;                                   /* 8 */ \
	BaseRequest* next_request;                   /* 8 */ \
	BaseRequest* next_anysource_request;         /* 8 */ \
	tasks::LocalTask* waiting_task               /* 8 */ ;

// we provide an init base request such that the whole request is initalized
#define INIT_BASE_REQUEST(REQ, PEER_RANK, CONTEXT_ID, TAG, BUF, BUFLEN, WAITINGTASK) \
	(REQ)->context_id = CONTEXT_ID; \
	(REQ)->peer_rank = PEER_RANK; \
	(REQ)->tag = TAG; \
	(REQ)->buflen = BUFLEN; \
	(REQ)->buf = BUF; \
	(REQ)->next_anysource_request = nullptr; \
	(REQ)->waiting_task = WAITINGTASK;

#define BASE_REQUEST_SIZE 128

//tinympi::BaseRequest
struct BaseRequest {
	BASE_REQUEST_MEMBERS
};

static_assert(std::is_standard_layout<BaseRequest>::value, "BaseRequest must be POD!");

#define ASSERT_BASE_REQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, atomic_state) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, provider_type) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, peer_rank) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, context_id) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, tag) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, buf) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, buflen) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, next_request) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, next_anysource_request) \
	VERIFY_COMPATIBILITY(BaseRequest, TYPE, waiting_task)

namespace baseRequest {

	namespace impl {
		inline bool set_request_to_waiting(BaseRequest* req) {
			atomic_state_type expected; expected.asInt = 0;
			expected.data.completion_status = completion_status_type::REQUEST_INPROGRESS;
			expected.data.waiting = 0;

			atomic_state_type desired  {completion_status_type::REQUEST_INPROGRESS, 1};
			int success = req->atomic_state.compare_exchange_strong(expected, desired);
			return success;
		}
	} // END NAMESPACE impl

	extern thread_local boost::pool<> request_pool_;

	inline BaseRequest* malloc() {
		return reinterpret_cast<BaseRequest*>(request_pool_.malloc());
	}
	inline void free(BaseRequest* req) {
		request_pool_.free(req);
	}

	// check whether any of the requests is completed
	inline bool is_terminated(BaseRequest* req) {
		BaseRequest* req_start = req;
		while (req != nullptr) {
			if (req->atomic_state.load().data.completion_status
					== completion_status_type::REQUEST_COMPLETED)
				return true;
			req = req->next_anysource_request;

			// if we are at the start again, no request is terminated
			if (req == req_start)
				return false;
		}
		return false;
	}
	inline void request_cancelled(BaseRequest* req) {
		using namespace impl;
		atomic_state_type old = req->atomic_state.exchange(
				{completion_status_type::REQUEST_CANCELLED, 0});
		if (old.data.waiting == 1) {
			req->waiting_task->dec_credit(1);
		}
	}
	inline void request_completed(BaseRequest* req) {
		using namespace impl;
		TMPI_LOG_LEVEL(1, "request " << req, "set to completed");
		atomic_state_type old = req->atomic_state.exchange(
				{completion_status_type::REQUEST_COMPLETED, 0});
		if (old.data.waiting == 1) {
			req->waiting_task->dec_credit(1);
		}
	}
	inline void init(BaseRequest* req, provider_types provider_type) {
		req->provider_type = provider_type;
		req->atomic_state.store({completion_status_type::REQUEST_INPROGRESS, 0},
				std::memory_order_release);
	}

	inline credit_t set_task_as_waiting(BaseRequest** req) {
		using namespace impl;
		credit_t res = 0;
		BaseRequest* breq = *req;
		res += set_request_to_waiting(breq) ? 1 : 0;
		if (breq->next_anysource_request != nullptr) {
			breq = breq->next_anysource_request;
			while (breq != *req) {
				res += set_request_to_waiting(breq) ? 1 : 0;
				breq = breq->next_anysource_request;
			}
		}
		return res;
	}

}

inline bool other_request_already_retrieved(BaseRequest* req);
void other_matched_anysource(BaseRequest* req, BaseRequest* tail_req);

} // END NAMESPACE impl
} // END NAMESPACE tinympi

#endif //_BASEREQUEST_H_
