// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _OFI_SENDREQUEST_H_635106f3226c46d3d114158785752466
#define _OFI_SENDREQUEST_H_635106f3226c46d3d114158785752466

#include <rdma/fi_domain.h>

#include "global.h"
#include "utility/error.h"

#include "ofi_endpoint.h"
#include "ofi_request.h"
#include "ofi_provider.h"

namespace tinympi {
namespace ofi {
namespace impl {

//tinympi::ofi::impl::send_request_state
enum class send_request_states {
	INJECT_INITIALIZED,
	NORMALSEND_INITIALIZED,
	SENDING,
	DONE
};

#define OFI_SEND_REQUEST_MEMBERS \
	OFI_REQUEST_MEMBERS \
	send_request_states req_state;

//tinympi::ofi::impl::OfiSendRequest
struct OfiSendRequest {
	OFI_SEND_REQUEST_MEMBERS
};
REQUESTS_CONVERTIBLE(OfiSendRequest, BaseRequest)
REQUESTS_CONVERTIBLE(OfiSendRequest, OfiRequest)

static_assert(std::is_standard_layout<OfiSendRequest>::value, "OfiSendRequest must be POD!");
ASSERT_OFI_REQUEST_COMPATIBLE(OfiSendRequest)
#define ASSERT_OFI_SENDREQUEST_COMPATIBLE(TYPE) \
	ASSERT_OFI_REQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(OfiSendRequest, TYPE, req_state)


namespace ofiSendRequest {
	namespace impl {

		// converts a state into a string
		inline std::string to_string(send_request_states state) {
			switch (state) {
			case send_request_states::INJECT_INITIALIZED:
				return "send_request_states::INITIALIZED";
			case send_request_states::NORMALSEND_INITIALIZED:
				return "send_request_states::NORMALSEND_INITIALIZED";
			case send_request_states::SENDING:
				return "send_request_states::SENDING";
			case send_request_states::DONE:
				return "send_request_states::DONE";
			default:
				return std::string("??? ") + std::to_string(static_cast<int>(state));
			}
		}

		inline void init(OfiSendRequest* req, send_request_states req_state_) {
			ofiRequest::init(reinterpret_cast<OfiRequest*>(req), ofi_request_types::OFI_SEND);
			req->req_state = req_state_;
		}

		// just set to done
		inline int sending_to_done(OfiSendRequest* req) {
			using namespace tinympi::impl;

			req->req_state = send_request_states::DONE;
			baseRequest::request_completed(TO_BaseRequest(req));

			return 1;
		}
	} // END NAMESPACE impl

	inline bool fill_status(OfiSendRequest*, TMPI_Status*) {
		return false;
	}

	// returns 1, if request is done
	inline int handle_poll(OfiSendRequest* req, fi_cq_tagged_entry&) {
		using namespace impl;

		TMPI_DASSERT_FATAL(req->req_state == send_request_states::SENDING,
				"unexpected request state (" << to_string(req->req_state) << ") at this place!",
				__FILE__, __LINE__);

		return sending_to_done(req);
	}

	inline OfiSendRequest* generate(void* buf, buflen_t buflen, tag_t tag,
			context_id_t context_id, const tasks::Task& dest_pos, OfiEndpoint* endpoint) {
		using namespace tinympi::impl;
		using namespace impl;

		uint64_t matchbits = ofiRequest::get_tag(context_id, tasks::get_threadunique_id(dest_pos),
				tag);
		OfiSendRequest* req = TO_OfiSendRequest(baseRequest::malloc());
		INIT_BASE_REQUEST( \
			req, \
			dest_pos.get_identifier().global_identifier, \
			context_id, \
			tag, \
			buf, \
			buflen, \
			tasks::get_current() \
		)
		req->endpoint = endpoint;

		if (buflen > provider_.get_max_inject_size()) {
			TMPI_LOG_LEVEL(3, "OfiSendRequest", "use normal send");

			// use tsend
			init(req, send_request_states::NORMALSEND_INITIALIZED);

			int res = -FI_EAGAIN;
			while (res == -FI_EAGAIN)
				res = fi_tsend(endpoint->get_ep(), buf, buflen, nullptr, endpoint->get_addr(),
						matchbits, &req->context_id);

			TMPI_ASSERT_FATAL(res == 0, "fi_tsend failed", __FILE__, __LINE__);

			req->req_state = send_request_states::SENDING;
			return req;

		} else {
			TMPI_LOG_LEVEL(3, "OfiSendRequest", "use injected send");


			init(req, send_request_states::INJECT_INITIALIZED);

			// use tinject
			int res = -FI_EAGAIN;
			while (res == -FI_EAGAIN)
				res = fi_tinject(endpoint->get_ep(), buf, buflen, endpoint->get_addr(), matchbits);

			TMPI_ASSERT_FATAL(res == 0, "fi_tinject failed", __FILE__, __LINE__);

			req->req_state = send_request_states::SENDING;
			sending_to_done(req);
			return req;

		}
	}

} // END NAMESPACE ofiSendRequest

} // END NAMESPACE impl
} // END NAMESPACE ofi
} // END NAMESPACE tinympi

#endif //_OFI_SENDREQUEST_H_
