// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_H_635106f3226c46d3d114158785752466
#define _VERBS_H_635106f3226c46d3d114158785752466

#include <vector>
#include <limits>
#include <type_traits>

#include "global.h"
#include "status.h"
#include "utility/error.h"
#include "tasks.h"

#include "verbs_queues.h"
#include "verbs_provider.h"

#include "verbs_eagersendrequest.h"
#include "verbs_eagerrecvrequest.h"
#include "verbs_rdvsendrequest.h"
#include "verbs_rdvrecvrequest.h"
#include "verbs_unknownrequest.h"
#include "verbs_mru.h"

#include "provider.h"

namespace tinympi {
namespace verbs {
namespace impl {

// if we have too much messages in the receive buffer, we need to explicitly send about free receive
// buffers
inline void send_explicit_piggyback(VerbsEndpoint* const ep) {
	LSB_RECORD(VERBS_SENDEXPLICITPIGGYBACK_STARTED);

	// get the piggyback message
	piggyback_t pstart;
	piggybackmsg_t pcnt;
	std::tie(pstart, pcnt) = ep->eager.get_explicit_piggyback_message();
	TMPI_LOG_LEVEL(2, "send_explicit_piggyback", "pstart = " << (int)pstart
			<< ", pcnt = " << (int)pcnt);

	// there can be zero, one or two sends.
	// > If there is no free entry, we do not send anything
	// > If there are multiple free entries, but not accross the border of the array, we send one
	//   message
	// > Otherwise we will send two messages
	piggybackmsg_t pcnt_1;
	if (pcnt == 0) {
		// case 0
		LSB_RECORD(VERBS_SENDEXPLICITPIGGYBACK_DONE);
		return;
	} else if (pstart + pcnt <= EAGER_CHMESSAGE_LIMIT)
		// case 1
		pcnt_1 = pcnt;
	else
		// case 2 (first free until the end of the buffer)
		pcnt_1 = EAGER_CHMESSAGE_LIMIT - pstart;

	// prepare send request
	ibv_sge sr_sg_lst = {
		reinterpret_cast<uint64_t>(VerbsProvider::Eager::get_piggyback_send_buffer()), // addr
		static_cast<uint32_t>(pcnt_1),                                                 // length
		VerbsProvider::Eager::get_piggyback_send_buffer_lkey()                         // lkey
	};

	// we overwrite the receive state array on the remote side
	ibv_send_wr sr_wr;
	memset(&sr_wr, 0, sizeof(ibv_send_wr));
	sr_wr.wr_id = 0;
	sr_wr.opcode = IBV_WR_RDMA_WRITE;
	sr_wr.send_flags = IBV_SEND_INLINE | IBV_SEND_SIGNALED;

	sr_wr.sg_list = &sr_sg_lst;
	sr_wr.num_sge = 1;
	sr_wr.next = nullptr;

	// we send into the remote status array of the peer
	const PeerInfo& pi = ep->eager.eager_get_remote_status_peer_info();
	sr_wr.wr.rdma.rkey = pi.rkey;
	sr_wr.wr.rdma.remote_addr = reinterpret_cast<uint64_t>(
			reinterpret_cast<uint8_t*>(pi.addr) + pstart);

	// post message
	ibv_send_wr* bad_wr;
	int res = ibv_post_send(ep->get_qp(), &sr_wr, &bad_wr);
	TMPI_LOG_LEVEL(2, "ibv_post_send", "piggyback (start = " << (int) pstart << "; pcnt = "
			<< (int) pcnt_1 << ")");
	TMPI_ASSERT_FATAL(res == 0, "ibv_post_send returned " << res, __FILE__, __LINE__);

	// if there is a second message, we repeat the procedure
	if (pcnt_1 != pcnt) {
		TMPI_LOG_LEVEL(2, "ibv_post_send", "piggyback (start = " << 0 << "; pcnt = "
				<< (int) pcnt - pcnt_1 << ")");

		// we send the remaining parts to the beginning of the remote status array of the peer
		sr_sg_lst.length = pcnt - pcnt_1;
		sr_wr.wr.rdma.remote_addr = pi.addr;
		int res = ibv_post_send(ep->get_qp(), &sr_wr, &bad_wr);
		TMPI_LOG_LEVEL(2, "ibv_post_send", "piggyback");
		TMPI_ASSERT_FATAL(res == 0, "ibv_post_send returned " << res, __FILE__, __LINE__);
	}
	LSB_RECORD(VERBS_SENDEXPLICITPIGGYBACK_DONE);

}

inline int poll_sready_channel(VerbsEndpoint* const ep) {
	using namespace impl::recvqueues;
	int pos;
	std::unique_lock<std::mutex> verbs_lock;

	int res = 0;
	while (ep->rdv.sready_poll_and_lock_recv_slots(ep->get_expected_rid(), pos, verbs_lock)) {
		TMPI_LOG_LEVEL(2, "poll_channels", "success: poll sready channel");
		LSB_RECORD(VERBS_POLLCHANNELS_SREADYPOLLANDLOCKRECVSLOTS);

		rid_t rid = ep->get_recv_rid();
		(void) rid;

		// if we received a message, the receive slot is used.
		const RdvSReadyMessage* message = ep->rdv.sready_get_recv_slot(pos);
		TMPI_LOG_LEVEL(4, "poll_channels", "   tag = " << message->tag << ", rid = " << (int)rid);

		// get the task
		tasks::Task& task = tasks::get_task(message->dest_id);

		// find request in waiting queue
		VerbsRequest* vreq = find_waiting_request<dequeue_request::yes>(message->src_id,
				message->tag, message->context_id, task);
		VerbsRdvRecvRequest* req = TO_VerbsRdvRecvRequest(vreq);
		if (req != nullptr) {
			TMPI_LOG_LEVEL(2, "poll_channels", "   found in waiting queue (req = " << req
					<< ", tag = " << (int)req->tag << ", context_id = " << (int)req->context_id
					<< ", src = " << req->peer_rank << ")");

			// when we get a message for a any sourcerequest, we need to delete in the other queues
			if (req->next_anysource_request != nullptr) {
				std::unique_lock<std::mutex> thread_lock(
						static_cast<tasks::LocalTask*>(&task)->thread_lock_);
				// required node + thread lock
				other_matched_anysource(req->next_anysource_request, TO_BaseRequest(req));
			}
			verbs_lock.unlock();

			// if this request exists, we notify the request (endpoint lock required)
			verbsRdvRecvRequest::retrieved_sready_for_waiting_request(req, message, ep);

			// free the recv slot. No lock is required for this
			ep->rdv.sready_free_recv_slot(pos);

			LSB_RECORD(VERBS_POLLCHANNELS_SREADYFOUNDDONE);

		} else {
			TMPI_LOG_LEVEL(2, "poll_channels", "   did not find in waiting queue");

			// if we did not find a request, generate a request and add to unexpected queue
			req = verbsRdvRecvRequest::retrieved_unexpected_rdv(pos, message, ep);
			vreq = TO_VerbsRequest(req);
			add_to_unexpected_queue(vreq, task);

			verbs_lock.unlock();
			LSB_RECORD(VERBS_POLLCHANNELS_SREADYNOTFOUNDDONE);
		}
		++res;
	}
	return res;
}

inline int poll_eager_channel(VerbsEndpoint* const ep) {
	using namespace impl::recvqueues;
	int pos;
	int res = 0;
	std::unique_lock<std::mutex> verbs_lock;

	// endpoint lock required for eager_poll_recv_slots, test_explicit_piggyback,
	// send_explicit_piggyback, retrieved_unexpected_eager, retrieved_eager_for_waiting_request
	while (ep->eager.eager_poll_and_lock_recv_slots(ep->get_expected_rid(), pos, verbs_lock)) {
		TMPI_LOG_LEVEL(2, "poll_channels", "success: poll eager channel");
		TMPI_LOG_LEVEL(2, "poll_channels", "   initialized EagerRecvRequest to WAITING_DATA");
		LSB_RECORD(VERBS_POLLCHANNELS_EAGERPOLLANDLOCK);

		// we only increase the rid AFTER we get the verbs_lock!
		rid_t rid = ep->get_recv_rid();
		(void) rid;

		// we search the message in the waiting queue (no lock required)
		const EagerMessage* message = ep->eager.eager_get_recv_slot(pos);
		TMPI_LOG_LEVEL(2, "poll_channels", "tag = " << message->tag << ", rid = " << message->rid);

		// get the task
		tasks::Task& task = tasks::get_task(message->dest_id);

		// find waiting request
		VerbsRequest* vreq = find_waiting_request<dequeue_request::yes>(
				message->src_id, message->tag, message->context_id, task);
		VerbsEagerRecvRequest* req = TO_VerbsEagerRecvRequest(vreq);
		if (req != nullptr) {
			TMPI_LOG_LEVEL(2, "poll_channels", "   found in waiting queue (req = " << req
					<< ", tag = " << (int)req->tag << ", context_id = " << (int)req->context_id
					<< ", src = " << req->peer_rank << ")");

			// when we get a message for a any sourcerequest, we need to delete in the other queues
			if (req->next_anysource_request != nullptr) {
				std::unique_lock<std::mutex> thread_lock(
						static_cast<tasks::LocalTask*>(&task)->thread_lock_);
				// required node + thread lock
				other_matched_anysource(req->next_anysource_request, TO_BaseRequest(req));
			}
			verbs_lock.unlock();

			// if this request exists, we notify it
			verbsEagerRecvRequest::retrieved_eager_for_waiting_request(req, message, ep);

			// we can free the recv slot, which is now consumed
			ep->eager.eager_free_recv_slot(pos);

			// when we receive a message, we probably need to send an explicit piggyback
			if (ep->eager.test_explicit_piggyback()) {
				// in this case we take the lock such that no other thread can send explicit
				// piggyback at the same time
				std::unique_lock<std::mutex> ep_lock(ep->lock);
				if (ep->eager.test_explicit_piggyback()) {
					send_explicit_piggyback(ep);
				}
				ep_lock.unlock();
			}
			LSB_RECORD(VERBS_POLLCHANNELS_EAGERFOUNDDONE);

		} else {
			TMPI_LOG_LEVEL(2, "poll_channels", "    did not find in queue");
			// if there is no message in the waiting queue, we add it to the unexpected queue
			req = verbsEagerRecvRequest::retrieved_unexpected_eager(pos, message, ep);
			vreq = TO_VerbsRequest(req);

			tasks::get_task(message->dest_id);
			TMPI_LOG_LEVEL(2, "poll_channels", "    add to unexpected queue");
			add_to_unexpected_queue(vreq, task);

			verbs_lock.unlock();

			LSB_RECORD(VERBS_POLLCHANNELS_EAGERNOTFOUNDDONE);
		}
		++res;
	}
	return res;
}
inline int poll_rready_channel(VerbsEndpoint* const ep) {
	using namespace impl::recvqueues;
	int pos;
	std::unique_lock<std::mutex> verbs_lock;

	int res = 0;
	while (ep->rdv.rready_poll_recv_slots(pos)) {

		TMPI_LOG_LEVEL(2, "poll_channels", "success: poll rready channel");
		LSB_RECORD(VERBS_POLLCHANNELS_RREADYPOLL);

		// we immediately free the recv slot
		ep->rdv.rready_free_recv_slot(pos);

		// if we got an rready, the send request is part of the message, so we just reconstruct
		// that pointer and forward the message to that request
		const RdvRReadyMessage* message = ep->rdv.rready_get_recv_slot(pos);
		VerbsRdvSendRequest* req = reinterpret_cast<VerbsRdvSendRequest*>(message->send_request);

		TMPI_LOG_LEVEL(2, "poll_channels", "   retrieved rready for request " << req << " (rid = "
				<< (int)req->rid << ")");

		verbsRdvSendRequest::retrieved_rready(req, message);
		LSB_RECORD(VERBS_POLLCHANNELS_RREADYPOLLDONE);
		++res;
	}
	return res;
}
// goes through all channels of an endpoint and detects whether new data is available
inline int poll_channels(VerbsEndpoint* const ep) {
	int res = 0;
	// no locks allowed

	// we then poll the sready_channel for a message with the correct request id. Polling is only
	// successful if we also get the verbs_lock.
	res += poll_sready_channel(ep);

	// and for eager messages
	res += poll_eager_channel(ep);

	// we search for rready messages
	res += poll_rready_channel(ep);

	return res;
}

//tinympi::verbs::impl::handle_wc
inline int handle_wc(VerbsEndpoint* const ep, const ibv_wc& wc) {
	// returns number of requests that have been finished
	// no lock allowed outside

	// if it is the completion event for an explicit piggyback, nothing has to be done
	if (wc.wr_id == 0 && wc.opcode != IBV_WC_RECV_RDMA_WITH_IMM)
		return 0;

	TMPI_LOG_LEVEL(1, "handle_wc", "Received work completion event.");

	// there are two types of work completion events: wc that are related to posted sends (wr_id!=0),
	// and received data (wr_id==0, further information in imm_data)
	if (wc.wr_id != 0) {
		TMPI_LOG_LEVEL(2, "handle_wc", "   work completion event with related request ("
				<< reinterpret_cast<void*>(wc.wr_id) << ")");

		// if wr_id != 0, it corresponds to the request pointer
		VerbsRequest* req = reinterpret_cast<VerbsRequest*>(wc.wr_id);

		// in any case, we just forward it to handle_poll of the correct request type
		int res;
		switch (req->verbs_request_type) {
		case verbs_request_types::VERBS_EAGERSEND:
			// NO LOCK required
			// retval OK
			res = verbsEagerSendRequest::handle_poll(TO_VerbsEagerSendRequest(req), wc);
			LSB_RECORD(VERBS_PROGRESS_WCEAGERSENDDONE);
			return res;

		// not needed
		// case verbs_request_types::VERBS_EAGERRECV:
			// return verbsEagerRecvRequest::handle_poll(TO_VerbsEagerRecvRequest(req), wc);

		case verbs_request_types::VERBS_RDVSEND:
			// NO LOCK required
			// retval ok
			res = verbsRdvSendRequest::handle_poll(TO_VerbsRdvSendRequest(req), wc);
			LSB_RECORD(VERBS_PROGRESS_WCRDVSENDDONE);
			return res;

		case verbs_request_types::VERBS_RDVRECV:
			// NO LOCK required
			// retval OK
			res = verbsRdvRecvRequest::handle_poll(TO_VerbsRdvRecvRequest(req), wc);
			LSB_RECORD(VERBS_PROGRESS_WCRDVRECVDONE);
			return res;

		default:
			TMPI_FATAL_ERROR("Unknown request type", __FILE__, __LINE__);
			exit(1);
		}

	} else { // if (wc.opcode == IBV_WC_RECV_RDMA_WITH_IMM) this is already checked above

#ifdef DEBUG
		ep->recvd_postwc();
#endif

		TMPI_LOG_LEVEL(2, "handle_wc", "   retrieved data");

		// if we received data, we can get the request from the endpoint.
		uint32_t pos = wc.imm_data;
		VerbsRdvRecvRequest* req = ep->rdv.get_waitingdata_request(pos);

		// we then just notify the request about the retrieved data
		// endpoint lock required
		int res = verbsRdvRecvRequest::retrieved_data(req);
		LSB_RECORD(VERBS_PROGRESS_RDVRETRIEVEDDATAHANDLED);
		// retval OK
		return res;
	}

	return 0;
}

} // END NAMESPACE impl

inline int fast_progress() {
	using namespace impl;
	if (provider_.countdown() == true) {
		int res = 0;
		int rec_nr = tasks::has_active_task() ? 10 : 11;
		for (global_id_t i = 0; i < num_endpoints_; ++i) {
			if (endpoints_[i] != nullptr) {

				// successful if we also get the verbs_lock.
				res += poll_sready_channel(endpoints_[i]);

				// we search for rready messages
				res += poll_rready_channel(endpoints_[i]);

			}
		}
		LSB_Set_Rparam_int("type", -res);
		LSB_Rec(rec_nr);
		LSB_Set_Rparam_int("type", 0);
	}
	return 0;
}

//tinympi::verbs::verbs_wrapper
struct verbs_wrapper {
private:
	// import some classes from the impl class to the namespace of the struct
	// we should not move them outside of impl, because they are not directly
	// availbale from the outside
	using VerbsEndpoint = verbs::impl::VerbsEndpoint;
	using VerbsRequest = verbs::impl::VerbsRequest;
	using BaseRequest = tinympi::impl::BaseRequest;
	using dispatch_methods = tinympi::impl::dispatch_methods;

	static int progress(VerbsEndpoint* ep);

public:
	using ProviderBaseRequest = verbs::impl::VerbsRequest;

	static int progress();
	static int isend(void*, buflen_t, tasks::Task*, tag_t, context_id_t, BaseRequest**);
	static int posted_request(void*, buflen_t, const tasks::Task*, tag_t, context_id_t,
			BaseRequest**);
	static bool recv_or_post_anysource(void*, buflen_t, tag_t, context_id_t, BaseRequest**);

	static void init();
	static void pre_init();
	static void finalize();
	static void post_finalize();

	static constexpr int max_request_size = max(
			max(sizeof(impl::VerbsEagerSendRequest), sizeof(impl::VerbsEagerRecvRequest)),
			max(sizeof(impl::VerbsRdvSendRequest),   sizeof(impl::VerbsRdvRecvRequest))
	);

	template<dispatch_methods dispatch_method, typename... Args>
	static int request_dispatch(VerbsRequest* req, Args... args) {
		TMPI_FATAL_ERROR("Invalid call!", __FILE__, __LINE__);
	}
};

template<>
inline int verbs_wrapper::request_dispatch<
		verbs_wrapper::dispatch_methods::posted_request_for_unexpected>(VerbsRequest* req, void* buf,
		buflen_t buflen) {
	// LOCK REQUIRED for endpoint
	using namespace impl;

	switch (req->verbs_request_type) {
	case verbs_request_types::VERBS_EAGERRECV:
		return verbsEagerRecvRequest::posted_request_for_unexpected_eager(
				TO_VerbsEagerRecvRequest(req), buf, buflen);

	case verbs_request_types::VERBS_RDVRECV:
		return verbsRdvRecvRequest::posted_request_for_unexpected_rdv(TO_VerbsRdvRecvRequest(req),
				buf, buflen);
	default:
		TMPI_FATAL_ERROR("unknown request type", __FILE__, __LINE__);
		return 0;
	}
}
// just dispatch to the right function.
template<>
inline int verbs_wrapper::request_dispatch<verbs_wrapper::dispatch_methods::fill_status>(
		VerbsRequest* req, TMPI_Status* status) {
	// no lock required
	using namespace impl;

	switch (req->verbs_request_type) {
	case verbs_request_types::VERBS_EAGERSEND:
	case verbs_request_types::VERBS_RDVSEND:
		return verbsSendRequest::fill_status(TO_VerbsSendRequest(req), status);

	case verbs_request_types::VERBS_EAGERRECV:
	case verbs_request_types::VERBS_RDVRECV:
		return verbsRecvRequest::fill_status(TO_VerbsRecvRequest(req), status);

	case verbs_request_types::VERBS_UNKNOWN:
		return verbsUnknownRequest::fill_status(TO_VerbsUnknownRequest(req), status);

	default:
		TMPI_FATAL_ERROR("unknown request type", __FILE__, __LINE__);
		return 0;
	}
}

template<>
inline int verbs_wrapper::request_dispatch<verbs_wrapper::dispatch_methods::delete_request>(
		VerbsRequest* vreq) {
	// no lock required
	using namespace impl;
	baseRequest::free(TO_BaseRequest(vreq));
	return TMPI_SUCCESS;
}
template<>
inline int verbs_wrapper::request_dispatch<verbs_wrapper::dispatch_methods::try_cancel_request>(
		VerbsRequest* vreq) {
	// LOCK REQUIRED FOR node queue (OK)
	using namespace impl;

	LSB_RECORD(VERBSWRAPPER_CANCELREQUEST_STARTED);

	// we currently only support cancellation of waiting requests (UnknownRequest)
	if (vreq->verbs_request_type != verbs_request_types::VERBS_UNKNOWN) {
		// otherwise the request has already been retrieved!
		LSB_RECORD(VERBSWRAPPER_CANCELREQUEST_DONE);
		return false;
	} else {
		int res = verbsUnknownRequest::try_cancel(TO_VerbsUnknownRequest(vreq));
		LSB_RECORD(VERBSWRAPPER_CANCELREQUEST_DONE);
		return res;
	}
}
// this function is called when a request was posted for a given source.
inline int verbs_wrapper::posted_request(void* buf, buflen_t buflen,
		const tasks::Task* src_descr, tag_t tag, context_id_t context_id, BaseRequest** req) {
	using namespace impl;
	LSB_RECORD(VERBSWRAPPER_POSTEDREQUEST_STARTED);

	// we need to lock the verbs queue
	// REQUIRED LOCK FOR find_unexpected_request, add_to_waiting_queue
	std::unique_lock<std::mutex> verbs_lock(tasks::get_current()->verbs_lock_);

	LSB_RECORD(VERBSWRAPPER_POSTEDREQUEST_ACQUIREDVERBSLOCK);

	// search request in unexpected queue and deque if found
	VerbsRequest* prev;
	*req = TO_BaseRequest(recvqueues::find_unexpected_request<recvqueues::dequeue_request::no>(
			src_descr->get_identifier().global_identifier, tag, context_id, prev));
	if (*req != nullptr) {

		// now we can dequeue
		recvqueues::dequeue_unexpected_request(TO_VerbsRequest(*req), prev);

		// if we dequeued the request, the verbs lock is not needed anymore, neither for rdv
		// nor for eager
		verbs_lock.unlock();

		// if we have found it, we forward our data to the request and retrieve the data
		int res = request_dispatch<dispatch_methods::posted_request_for_unexpected>(
				TO_VerbsRequest(*req), buf, buflen);
		LSB_RECORD(VERBSWRAPPER_POSTEDREQUEST_DISPATCHED);
		return res;

	} else {

		// otherwise, we enqueue a new request to the waiting queue
		*req = TO_BaseRequest(verbsUnknownRequest::generate_waiting_request(*src_descr, tag,
				context_id, buf, buflen));

		recvqueues::add_to_waiting_queue(TO_VerbsRequest(*req));

		verbs_lock.unlock();

// 		mrulist::prioritize_endpoint(TO_VerbsRequest(*req)->endpoint);
		LSB_RECORD(VERBSWRAPPER_POSTEDREQUEST_ADDEDTOWAITINGQUEUE);

		return TMPI_SUCCESS;
	}
}
// this function is called when a request was posted with ANY_SOURCE as source
inline bool verbs_wrapper::recv_or_post_anysource(void* buf, buflen_t buflen, tag_t tag,
		context_id_t context_id, BaseRequest** req) {
	// LOCK REQUIRED FOR node queue, for find_unexpected_anysource_request, add_to_waiting_queue (OK)
	LSB_RECORD(VERBSWRAPPER_RECVORPOSTANYSOURCE_STARTED);

	using namespace impl;

	VerbsRequest* prev;
	VerbsRequest* vreq = recvqueues::find_unexpected_anysource_request
			<recvqueues::dequeue_request::yes>(tag, context_id, prev);
	if (vreq != nullptr) {
		TMPI_LOG_LEVEL(2, "NodeWrapper", "found unexpected request");

		// add old any source to new request
		vreq->next_anysource_request = *req;
		*req = TO_BaseRequest(vreq);
		LSB_RECORD(VERBSWRAPPER_RECVORPOSTANYSOURCE_FOUNDUNEXPECTED);
		return true;

	} else {
		TMPI_LOG_LEVEL(2, "NodeWrapper", "did not find unexpected request");

		// otherwise, we enqueue a new request to the waiting queue
		vreq = TO_VerbsRequest(verbsUnknownRequest::generate_anysource_waiting_request(tag,
				context_id, buf, buflen));
		recvqueues::add_to_waiting_queue(vreq);

		// add old any source to new request
		vreq->next_anysource_request = *req;
		*req = TO_BaseRequest(vreq);

		LSB_RECORD(VERBSWRAPPER_RECVORPOSTANYSOURCE_ADDEDTOWAITING);
		return false;
	}
}
// do progress for one endpoint
inline int verbs_wrapper::progress(VerbsEndpoint* ep) {
	// NO LOCKS ALLOWED
	int res;

	// then we poll the completion queue
	ibv_wc wc[NUMBER_OF_WC_PER_POLL];
	res = ibv_poll_cq(ep->get_cq(), NUMBER_OF_WC_PER_POLL, wc);
	TMPI_ASSERT_FATAL(res != -1, "ibv_poll_cq() returned " << res, __FILE__, __LINE__);

	for (int i = 0; i < res; ++i) {
		TMPI_ASSERT_FATAL(wc[i].status == IBV_WC_SUCCESS, "wc[i].wr_id "
				<< reinterpret_cast<void*>(wc[i].wr_id) << " failed with status \""
				<< ibv_wc_status_str(wc[i].status) << "\" (" << wc[i].status << ")"
				, __FILE__, __LINE__);
		// if there is some entry in the completion queue which succeeded, we handle it
		LSB_RECORD(VERBS_PROGRESS_POLLEDWC);
		handle_wc(ep, wc[i]);
		LSB_RECORD(VERBS_PROGRESS_WCHANDLED);
	}
	if (tasks::has_active_task())
		return res;

	// we poll all channels
	// if polling the channels was successful, there is no reason to spend time polling the cq
	res += poll_channels(ep);
	return res;
}

// just get one endpoint from the mrulist and forward
inline int verbs_wrapper::progress() {
	using namespace impl;
	// NO LOCKS ALLOWED
	int res = 0;
	for (global_id_t i = 0; i < num_endpoints_; ++i) {
		if (endpoints_[i] != nullptr) {
			res += progress(endpoints_[i]);
		}
		if (tasks::has_active_task())
			return res;
	}
// 	return progress(mrulist::get_endpoint());
	return res;
}

inline int verbs_wrapper::isend(void* buf, buflen_t buflen, tasks::Task* dest_descr,
		tag_t tag, context_id_t context_id, BaseRequest** reqptr) {
	// no endpoint lock allowed. endpoint is locked in generate
	using namespace impl;
	LSB_RECORD(VERBSWRAPPER_ISEND_STARTED);

	// get the endpoint of the respective node
	VerbsEndpoint* endpoint = get_endpoint(dest_descr->get_identifier().local.node_id);
// 	mrulist::prioritize_endpoint(endpoint);

	// if the data is small enough to be sent with eager, we generate a EagerSendRequest
	if (buflen < EAGER_MESSAGE_THRESH) {
		TMPI_LOG_LEVEL(1, "isend", "try to generate eager send request");
		VerbsEagerSendRequest* req = verbsEagerSendRequest::generate(buf, buflen, tag,
				context_id, *dest_descr, endpoint);
		*reqptr = TO_BaseRequest(req);

		if (*reqptr != nullptr) {
			LSB_RECORD(VERBSWRAPPER_ISEND_GENERATEDEAGERSENDREQUEST);
			return TMPI_SUCCESS;
		}

	}

	// otherwise we generate an RdvSendRequest (this can happen when either we have no more eager
	// send slots, or the message is too large)
	TMPI_LOG_LEVEL(1, "isend", "generate rdv send request");
	VerbsRdvSendRequest* req = verbsRdvSendRequest::generate(buf, buflen, tag, context_id,
			*dest_descr, endpoint);
	*reqptr = TO_BaseRequest(req);
	TMPI_DASSERT_FATAL(reqptr != nullptr, "could not create RdvSendRequest", __FILE__, __LINE__);
	LSB_RECORD(VERBSWRAPPER_ISEND_GENERATEDRDVSENDREQUEST);

	return TMPI_SUCCESS;
}

inline void verbs_wrapper::init() {
	using namespace impl;
	provider_.post_init();
	mrulist::init(endpoints_, num_endpoints_);
}
inline void verbs_wrapper::pre_init() {
	using namespace impl;

	provider_.init();
	VerbsProvider::setup_endpoints();
}
inline void verbs_wrapper::finalize() {
}
inline void verbs_wrapper::post_finalize() {
	using namespace impl;

	VerbsProvider::destroy_endpoints();
	TMPI_LOG("VerbsProvider","Destroyed endpoints");
	provider_.finalize();
}


} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif //_VERBS_H_
