// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#include <cstring>
#include <functional>

#include <malloc_overwrite.h>

#include "communicator.h"

#include "utility/external.h"
#include "utility/error.h"

#include "verbs_provider.h"

#include "verbs_rdvrreadymessage.h"
#include "verbs_rdvsreadymessage.h"
#include "verbs_eagermessage.h"

namespace tinympi {
namespace verbs {
namespace impl {

VerbsProvider provider_;
VerbsEndpoint** endpoints_;
global_id_t num_endpoints_;

thread_local bool VerbsProvider::recursive_ibvregmr_call = false;

void VerbsProvider::init() {

	// open device list
	int num_devices;
	ibv_device** dev_list = ibv_get_device_list(&num_devices);
	TMPI_ASSERT_FATAL(dev_list != nullptr, "ibv_get_device_list() returned 0", __FILE__, __LINE__);
	TMPI_ASSERT_FATAL(num_devices > 0, "ibv_get_device_list() did not find anz device!", __FILE__,
			__LINE__);

	// use first device and open
	ibv_device* device = dev_list[0];
	TMPI_ASSERT_FATAL(dev_list != nullptr, "detected device does not exist!", __FILE__, __LINE__);
	context_ = ibv_open_device(device);
	TMPI_ASSERT_FATAL(context_ != nullptr, "ibv_open_device() returned 0", __FILE__, __LINE__);

	// close device list
	ibv_free_device_list(dev_list);

	// get device name and print to log
	const char* dev_name = ibv_get_device_name(device);
	TMPI_ASSERT_FATAL(dev_name != nullptr, "couldn't get device name", __FILE__, __LINE__);
	TMPI_LOG_LEVEL(0, "device name: ", dev_name);

	// allocate protection domain (for AH, QP, MR, MW and SRQ)
	protection_domain_ = ibv_alloc_pd(context_);
	TMPI_ASSERT_FATAL(protection_domain_ != nullptr, "couldn't create protection domain", __FILE__,
			__LINE__);

	RdvProvider::init(get_pd());
	EagerProvider::init(get_pd());

	set_malloc_hook([](void* ptr, size_t sz) {
		provider_.malloc_hook(ptr, sz);
	});
}
void VerbsProvider::post_init() {
	EagerProvider::post_init();
	RdvProvider::post_init();
}
void VerbsProvider::finalize() {
	int res;
	EagerProvider::finalize();
	RdvProvider::finalize();

	memory_regions_.clear();

	res = ibv_dealloc_pd(protection_domain_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dealloc_pd returned " << res, __FILE__, __LINE__);

	res = ibv_close_device(context_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_close_device returned " << res, __FILE__, __LINE__);
}

void VerbsProvider::setup_endpoints() {
	int res;

	num_endpoints_ = ENDPOINTSPERNODE*utility::external::get_numnodes();
	global_id_t node_rank = utility::external::get_noderank();

	endpoints_ = new VerbsEndpoint*[num_endpoints_];

	uint16_t* lids = new uint16_t[num_endpoints_];
	uint32_t* qp_nums = new uint32_t[num_endpoints_];
	constexpr int num_mr = 3; // we have three channels (rready, sready and eager)
	uint32_t* recv_rkeys = new uint32_t[num_mr*num_endpoints_];
	uint64_t* recv_addrs = new uint64_t[num_mr*num_endpoints_];
	uint32_t* remote_rkeys = new uint32_t[num_mr*num_endpoints_];
	uint64_t* remote_addrs = new uint64_t[num_mr*num_endpoints_];

	ibv_port_attr port_attr;
	res = ibv_query_port(provider_.context_, 1, &port_attr);
	TMPI_ASSERT_FATAL(res == 0, "ibv_query_port() returned " << res, __FILE__, __LINE__);

	for (global_id_t i = 0; i < num_endpoints_; ++i) {
		if (i % (num_endpoints_/ENDPOINTSPERNODE) != node_rank) {
			VerbsEndpoint* ep = new VerbsEndpoint;

			ibv_cq* cq = ibv_create_cq(provider_.context_, 8192, nullptr, nullptr, 0);
			TMPI_ASSERT_FATAL(cq != nullptr, "could not create completion queue", __FILE__, __LINE__);
			ibv_qp_init_attr qp_init_attr;
			memset(&qp_init_attr, 0, sizeof(ibv_qp_init_attr));
			qp_init_attr.qp_context = provider_.context_;
			qp_init_attr.send_cq = cq;
			qp_init_attr.recv_cq = cq;
			qp_init_attr.cap.max_send_wr = 512;
			qp_init_attr.cap.max_recv_wr = 512;
			qp_init_attr.cap.max_send_sge = 1;
			qp_init_attr.cap.max_recv_sge = 1;
			qp_init_attr.cap.max_inline_data = provider_.max_inline_data_;
			qp_init_attr.qp_type = IBV_QPT_RC;
			qp_init_attr.sq_sig_all = 0;

			ibv_qp* qp = ibv_create_qp(provider_.protection_domain_, &qp_init_attr);
			TMPI_ASSERT_FATAL(qp != nullptr, "could not create qp", __FILE__, __LINE__);
			qp_nums[i] = qp->qp_num;
			lids[i] = port_attr.lid;

			ep->setup(cq, qp, provider_.protection_domain_);

			ep->eager.eager_get_recv_slots_info(recv_rkeys[i*num_mr  ], recv_addrs[i*num_mr  ]);
			ep->rdv.rready_get_recv_slots_info (recv_rkeys[i*num_mr+1], recv_addrs[i*num_mr+1]);
			ep->rdv.sready_get_recv_slots_info (recv_rkeys[i*num_mr+2], recv_addrs[i*num_mr+2]);

			ep->eager.eager_get_remote_info(remote_rkeys[i*num_mr  ], remote_addrs[i*num_mr  ]);
			ep->rdv.rready_get_remote_info (remote_rkeys[i*num_mr+1], remote_addrs[i*num_mr+1]);
			ep->rdv.sready_get_remote_info (remote_rkeys[i*num_mr+2], remote_addrs[i*num_mr+2]);

			endpoints_[i] = ep;
		} else {
			endpoints_[i] = nullptr;
		}
	}

	// then we should communicate the lids, the queue pair numbers, and memory all memory regions
	uint16_t* lids_recv = new uint16_t[num_endpoints_];
	uint32_t* qp_nums_recv = new uint32_t[num_endpoints_];
	uint32_t* recv_rkeys_recv = new uint32_t[3*num_endpoints_];
	uint64_t* recv_addrs_recv = new uint64_t[3*num_endpoints_];
	uint32_t* remote_rkeys_recv = new uint32_t[3*num_endpoints_];
	uint64_t* remote_addrs_recv = new uint64_t[3*num_endpoints_];

	for (unsigned i = 0; i < ENDPOINTSPERNODE; ++i) {
		res = utility::external::all2all(lids+num_endpoints_/ENDPOINTSPERNODE*i, 1, MPI_SHORT,
				lids_recv+num_endpoints_/ENDPOINTSPERNODE*i, 1, MPI_SHORT, MPI_COMM_WORLD);
		TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "MPI_Alltoall returned " << res, __FILE__, __LINE__);

		res = utility::external::all2all(qp_nums+num_endpoints_/ENDPOINTSPERNODE*i, 1, MPI_INT,
				qp_nums_recv+num_endpoints_/ENDPOINTSPERNODE*i, 1, MPI_INT, MPI_COMM_WORLD);
		TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "MPI_Alltoall returned " << res, __FILE__, __LINE__);

		res = utility::external::all2all(recv_rkeys+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_INT,
				recv_rkeys_recv+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_INT, MPI_COMM_WORLD);
		TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "MPI_Alltoall returned " << res, __FILE__, __LINE__);

		res = utility::external::all2all(recv_addrs+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_LONG,
				recv_addrs_recv+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_LONG, MPI_COMM_WORLD);
		TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "MPI_Alltoall returned " << res, __FILE__, __LINE__);

		res = utility::external::all2all(remote_rkeys+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_INT,
				remote_rkeys_recv+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_INT, MPI_COMM_WORLD);
		TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "MPI_Alltoall returned " << res, __FILE__, __LINE__);

		res = utility::external::all2all(remote_addrs+3*num_endpoints_/ENDPOINTSPERNODE*i, 3,
				MPI_LONG, remote_addrs_recv+3*num_endpoints_/ENDPOINTSPERNODE*i, 3, MPI_LONG,
				MPI_COMM_WORLD);
		TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "MPI_Alltoall returned " << res, __FILE__, __LINE__);
	}

	delete[] lids;
	delete[] qp_nums;
	delete[] recv_rkeys;
	delete[] recv_addrs;
	delete[] remote_rkeys;
	delete[] remote_addrs;

	for (global_id_t i = 0; i < num_endpoints_; ++i) {
		if (endpoints_[i] != nullptr) {
			pair_endpoints(endpoints_[i], lids_recv[i], qp_nums_recv[i], &recv_rkeys_recv[num_mr*i],
					&recv_addrs_recv[num_mr*i], &remote_rkeys_recv[num_mr*i],
					&remote_addrs_recv[num_mr*i]);
		}
	}

	delete[] lids_recv;
	delete[] qp_nums_recv;
	delete[] recv_rkeys_recv;
	delete[] recv_addrs_recv;
	delete[] remote_rkeys_recv;
	delete[] remote_addrs_recv;

	TMPI_LOG_LEVEL(0, "setup", "endpoints initialized");
}

void VerbsProvider::pair_endpoints(VerbsEndpoint* ep, uint16_t lid, uint32_t qp_num,
		uint32_t* recv_rkeys, uint64_t* recv_addrs, uint32_t* remote_rkeys, uint64_t* remote_addrs) {
	// to init state
	ibv_qp_attr qp_attr_to_init;
	memset(&qp_attr_to_init, 0, sizeof(ibv_qp_attr));
	qp_attr_to_init.qp_state = IBV_QPS_INIT;
	qp_attr_to_init.pkey_index = 0;
	qp_attr_to_init.qp_access_flags = IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_WRITE |
			IBV_ACCESS_LOCAL_WRITE;
	qp_attr_to_init.port_num = 1;
	int res = ibv_modify_qp(ep->get_qp(), &qp_attr_to_init,
			IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_ACCESS_FLAGS | IBV_QP_PORT);
	TMPI_ASSERT_FATAL(res == 0, "ibv_modify_qp() returned " << res, __FILE__, __LINE__);

	// to rtr state
	ibv_qp_attr qp_attr_to_rtr;
	memset(&qp_attr_to_rtr, 0, sizeof(ibv_qp_attr));
	qp_attr_to_rtr.qp_state = IBV_QPS_RTR;
	qp_attr_to_rtr.ah_attr.dlid = lid;
	qp_attr_to_rtr.ah_attr.sl = 0;
	qp_attr_to_rtr.ah_attr.src_path_bits = 0;
	qp_attr_to_rtr.ah_attr.static_rate = 0;
	qp_attr_to_rtr.ah_attr.is_global = 0;
	qp_attr_to_rtr.ah_attr.port_num = 1;
	qp_attr_to_rtr.rq_psn = 0;
	qp_attr_to_rtr.max_dest_rd_atomic = 4;
	qp_attr_to_rtr.min_rnr_timer = 12;
	qp_attr_to_rtr.dest_qp_num = qp_num;
	qp_attr_to_rtr.path_mtu = IBV_MTU_2048;
	res = ibv_modify_qp(ep->get_qp(), &qp_attr_to_rtr,
		IBV_QP_STATE | IBV_QP_AV | IBV_QP_RQ_PSN | IBV_QP_MAX_DEST_RD_ATOMIC
		| IBV_QP_MIN_RNR_TIMER | IBV_QP_DEST_QPN | IBV_QP_PATH_MTU);
	TMPI_ASSERT_FATAL(res == 0, "ibv_modify_qp() returned " << res, __FILE__, __LINE__);

	ibv_qp_attr qp_attr_to_rts;
	memset(&qp_attr_to_rts, 0, sizeof(ibv_qp_attr));
	qp_attr_to_rts.qp_state = IBV_QPS_RTS;
	qp_attr_to_rts.sq_psn = 0;
	qp_attr_to_rts.timeout = 1;
	qp_attr_to_rts.retry_cnt = 7;
	qp_attr_to_rts.rnr_retry = 7;
	qp_attr_to_rts.max_rd_atomic = 4;
	res = ibv_modify_qp(ep->get_qp(), &qp_attr_to_rts,
		IBV_QP_STATE | IBV_QP_SQ_PSN | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT
		| IBV_QP_RNR_RETRY | IBV_QP_MAX_QP_RD_ATOMIC);
	TMPI_ASSERT_FATAL(res == 0, "ibv_modify_qp() returned " << res, __FILE__, __LINE__);

	// then set peer memory information
	ep->eager.eager_set_peer_info(recv_rkeys[0], recv_addrs[0], remote_rkeys[0], remote_addrs[0]);
	ep->rdv.rready_set_peer_info (recv_rkeys[1], recv_addrs[1], remote_rkeys[1], remote_addrs[1]);
	ep->rdv.sready_set_peer_info (recv_rkeys[2], recv_addrs[2], remote_rkeys[2], remote_addrs[2]);
}

void VerbsProvider::destroy_endpoints() {

	int res;
	for (global_id_t i = 0; i < num_endpoints_; ++i) {
		if (endpoints_[i] != nullptr) {
			VerbsEndpoint* ep = endpoints_[i];

			ep->destroy();

			res = ibv_destroy_qp(ep->get_qp());
			TMPI_ASSERT_FATAL(res == 0, "ibv_destroy_qp returned " << res, __FILE__, __LINE__);

			res = ibv_destroy_cq(ep->get_cq());
			TMPI_ASSERT_FATAL(res == 0, "ibv_destroy_cq returned " << res, __FILE__, __LINE__);
		}
	}
	delete[] endpoints_;
}

} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi
