// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_RDVPROVIDER_H_635106f3226c46d3d114158785752466
#define _VERBS_RDVPROVIDER_H_635106f3226c46d3d114158785752466

#include <atomic>

#include <malloc.h>

#include "global.h"
#include "debug/debugger.h"
#include "utility/error.h"
#include "tasks.h"

#include "verbs_config.h"

#include "verbs_rdvsreadymessage.h"
#include "verbs_rdvrreadymessage.h"

namespace tinympi {
namespace verbs {
namespace impl {

class VerbsRdvSendRequest;

//tinympi::verbs::impl::VerbsProvider
class RdvProvider {
public:
	//////////////////////////////
	// initialization functions
	static void init(ibv_pd* pd);
	static void post_init();
	static void finalize();

	//////////////////////////////
	// sready send slots functions
	static void sready_free_send_slot(int);
	static std::pair<int, RdvSReadyMessage*> sready_get_send_slot();
	static uint32_t sready_get_send_slots_lkey();

	//////////////////////////////
	// rready send slots functions
	static void rready_free_send_slot(int);
	static std::pair<int, RdvRReadyMessage*> rready_get_send_slot();
	static uint32_t rready_get_send_slots_lkey();

private:
	//////////////////////////////
	// sready send slots
	// we have a limited amount of sready send slots that can be used by
	// all endpoints
	static RdvSReadyMessage* sready_send_slots_;
	// and the memory region
	static ibv_mr* sready_send_slots_mr_;

	// next send slot to be used
	static thread_local uint_fast16_t sready_send_slots_head_;
	// send slot base
	static thread_local uint_fast16_t sready_send_slots_start_;

	//////////////////////////////
	// rready send slots
	// we have a limited amount of rready send slots that can be used by
	// all endpoints
	static RdvRReadyMessage* rready_send_slots_;
	// and the memory region
	static ibv_mr* rready_send_slots_mr_;

	// next send slot to be used
	static thread_local uint_fast16_t rready_send_slots_head_;
	// send slot base
	static thread_local uint_fast16_t rready_send_slots_start_;

};

//////////////////////////////
// INITIALIZATION FUNCTIONS
// called by only one thread
inline void RdvProvider::init(ibv_pd* pd) {
	//////////////////////////////
	// sready send slots
	sready_send_slots_ = (RdvSReadyMessage*) memalign(64,
			env::num_threads*RDV_MESSAGE_LIMIT*sizeof(RdvSReadyMessage));
	for (unsigned i = 0; i < env::num_threads*RDV_MESSAGE_LIMIT; ++i)
		sready_send_slots_[i].status = sready_message_status::free;
	sready_send_slots_mr_ = ibv_reg_mr(pd, sready_send_slots_,
			env::num_threads*RDV_MESSAGE_LIMIT*sizeof(RdvSReadyMessage),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(sready_send_slots_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);
	sready_send_slots_head_ = 0;

	//////////////////////////////
	// rready send slots
	rready_send_slots_ = (RdvRReadyMessage*) memalign(64,
			env::num_threads*RDV_MESSAGE_LIMIT*sizeof(RdvRReadyMessage));
	for (unsigned i = 0; i < env::num_threads*RDV_MESSAGE_LIMIT; ++i)
		rready_send_slots_[i].status = rready_message_status::free;
	rready_send_slots_mr_ = ibv_reg_mr(pd, rready_send_slots_,
			env::num_threads*RDV_MESSAGE_LIMIT*sizeof(RdvRReadyMessage),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(rready_send_slots_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);
	rready_send_slots_head_ = 0;

}
// called by all threads
inline void RdvProvider::post_init() {
	sready_send_slots_start_ = threads::impl::self->rank * RDV_MESSAGE_LIMIT;
	rready_send_slots_start_ = threads::impl::self->rank * RDV_MESSAGE_LIMIT;
}
// called by only one thread
inline void RdvProvider::finalize() {
	int res;

	//////////////////////////////
	// sready send slots
	res = ibv_dereg_mr(sready_send_slots_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);
	free(sready_send_slots_);

	//////////////////////////////
	// rready send slots
	res = ibv_dereg_mr(rready_send_slots_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);
	free(rready_send_slots_);

}

//////////////////////////////
// SREADY SEND SLOTS FUNCTIONS
// free a send slot
// called by only one thread
inline void RdvProvider::sready_free_send_slot(int i) {
	TMPI_DASSERT_FATAL(sready_send_slots_[i].status != sready_message_status::free,
			"double freed send slot", __FILE__, __LINE__);

	sready_send_slots_[i].status = sready_message_status::free;
}
// get a free send slot
inline std::pair<int, RdvSReadyMessage*> RdvProvider::sready_get_send_slot() {

	while (sready_send_slots_[sready_send_slots_start_ +
			(sready_send_slots_head_ & (RDV_MESSAGE_LIMIT - 1))].status
			!= sready_message_status::free) {
		sready_send_slots_head_ += 1;
	}
	auto my_slot = sready_send_slots_start_ + (sready_send_slots_head_ & (RDV_MESSAGE_LIMIT - 1));
	sready_send_slots_head_ += 1;

	// this thread sets the slot to retrieved
	sready_send_slots_[my_slot].status = sready_message_status::retrieved;
	return {my_slot, &sready_send_slots_[my_slot]};
}
// trivially thread save
inline uint32_t RdvProvider::sready_get_send_slots_lkey() {
	return sready_send_slots_mr_->lkey;
}

//////////////////////////////
// RREADY SEND SLOTS FUNCTIONS
// free a send slot
// called by only one thread
inline void RdvProvider::rready_free_send_slot(int i) {
	TMPI_DASSERT_FATAL(rready_send_slots_[i].status != rready_message_status::free,
			"double freed send slot", __FILE__, __LINE__);

	rready_send_slots_[i].status = rready_message_status::free;
}
// get a free send slot
inline std::pair<int, RdvRReadyMessage*> RdvProvider::rready_get_send_slot() {

	// we need to check whether the slots are full. This is the case when the current slot is
	// reserved. Pay attention to thread safety!
	while (rready_send_slots_[rready_send_slots_start_ +
			(rready_send_slots_head_ & (RDV_MESSAGE_LIMIT - 1))].status
			!= rready_message_status::free) {
		rready_send_slots_head_ += 1;
	}
	auto my_slot = rready_send_slots_start_ + (rready_send_slots_head_ & (RDV_MESSAGE_LIMIT - 1));
	rready_send_slots_head_ += 1;

	// this thread sets the slot to retrieved
	rready_send_slots_[my_slot].status = rready_message_status::retrieved;

	return {my_slot, &rready_send_slots_[my_slot]};
}
// trivially thread safe
inline uint32_t RdvProvider::rready_get_send_slots_lkey() {
	return rready_send_slots_mr_->lkey;
}


} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif // _VERBS_RDVPROVIDER_H
