// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_RDVRECVREQUEST_H_635106f3226c46d3d114158785752466
#define _VERBS_RDVRECVREQUEST_H_635106f3226c46d3d114158785752466

#include <memory>

#include "global.h"

#include "verbs_endpoint.h"
#include "verbs_request.h"
#include "verbs_provider.h"
#include "verbs_config.h"
#include "verbs_recvrequest.h"
#include "verbs_mrwrapper.h"

namespace tinympi {
namespace verbs {
namespace impl {

//tinympi::verbs::impl::recv_request_state
enum class rdv_recv_request_common_states {
	INITIALIZED,
	WAITING_SREADY,
	UNEXPECTED_REQUEST,
	SENDING_RREADY,
	DONE // =RECVD_DATA
};

#define VERBS_RDVRECV_REQUEST_MEMBERS \
	VERBS_RECV_REQUEST_MEMBERS \
	rdv_recv_request_common_states req_common_state; \
	std::atomic<uint8_t> req_rready_state; \
	int send_rready_slot_pos; \
	int rem_rready_message_pos; \
	std::shared_ptr<mr_wrapper> mr;


//tinympi::verbs::impl::VerbsRdvRecvRequest
struct VerbsRdvRecvRequest {
	VERBS_RDVRECV_REQUEST_MEMBERS
};
static_assert(std::is_standard_layout<VerbsRdvRecvRequest>::value,
		"VerbsRdvRecvRequest must be POD!");

REQUESTS_CONVERTIBLE(VerbsRdvRecvRequest, VerbsRecvRequest)
REQUESTS_CONVERTIBLE(VerbsRdvRecvRequest, VerbsRequest)
REQUESTS_CONVERTIBLE(VerbsRdvRecvRequest, BaseRequest)

ASSERT_VERBS_RECVREQUEST_COMPATIBLE(VerbsRdvRecvRequest)

#define ASSERT_VERBS_RDVRECVREQUEST_COMPATIBLE(TYPE) \
	ASSERT_VERBS_RECVREQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(VerbsRdvRecvRequest, TYPE, req_common_state) \
	VERIFY_COMPATIBILITY(VerbsRdvRecvRequest, TYPE, req_rready_state) \
	VERIFY_COMPATIBILITY(VerbsRdvRecvRequest, TYPE, send_rready_slot_pos) \
	VERIFY_COMPATIBILITY(VerbsRdvRecvRequest, TYPE, rem_rready_message_pos) \
	VERIFY_COMPATIBILITY(VerbsRdvRecvRequest, TYPE, mr)

namespace verbsRdvRecvRequest {
	namespace impl {
		// converts a state into a string
		inline std::string to_string(rdv_recv_request_common_states state) {
			// no lock required
			switch (state) {
			case rdv_recv_request_common_states::INITIALIZED:
				return "rdv_recv_request_common_states::INITIALIZED";
			case rdv_recv_request_common_states::WAITING_SREADY:
				return "rdv_recv_request_common_states::WAITING_SREADY";
			case rdv_recv_request_common_states::UNEXPECTED_REQUEST:
				return "rdv_recv_request_common_states::UNEXPECTED_REQUEST";
			case rdv_recv_request_common_states::SENDING_RREADY:
				return "rdv_recv_request_common_states::SENDING_RREADY";
			case rdv_recv_request_common_states::DONE:
				return "rdv_recv_request_common_states::DONE";
			}
			return std::string("??? ") + std::to_string(static_cast<int>(state));
		}

		// initialize a new request
		inline void init(VerbsRdvRecvRequest* const req) {
			// no lock required
			verbsRecvRequest::init(TO_VerbsRecvRequest(req), verbs_request_types::VERBS_RDVRECV);
			std::memset(&req->mr, 0, sizeof(req->mr));
			req->req_common_state = rdv_recv_request_common_states::INITIALIZED;
		}

		/////////////////////////////
		// secondary state changes
		inline int waitingsready_to_sendingrready(VerbsRdvRecvRequest* const req,
				const RdvSReadyMessage* const message) {
			// endpoint lock required, e.g, for add_waiting_request, rready_reserve_remote_slot
			LSB_RECORD(VERBS_RDVWAITINGSREADYTOSENDIGNRREADY_STARTED);

			// when the sready message arrives, we can send the rready as soon as we are ready
			TMPI_ASSERT_FATAL(message->buflen <= req->buflen, "overflow error: "
					<< "recvd buffer has size " << message->buflen
					<< ", expected size: " << req->buflen, __FILE__, __LINE__);
			// update size, tag and source
			req->buflen = message->buflen;

			// update our tag (may be ANY_TAG)
			req->tag = message->tag;

			// we need to register the recv buffer. We cannot do that earlier,
			// because only at this point we know that we will receive data with the rendezvous
			// protocol
			if (req->buflen > 0) {
				req->mr = provider_.register_memory(req->buf, req->buflen);
			}

			int res;

			// find a send slot
			RdvRReadyMessage* rready_message;
			// this also sets the status to reserved
			std::tie(req->send_rready_slot_pos, rready_message)
					= VerbsProvider::Rdv::rready_get_send_slot();
			TMPI_ASSERT_FATAL(rready_message != nullptr, "too few rready messages! increase "
					"RDV_MESSAGE_LIMIT.", __FILE__, __LINE__);

			uint32_t identifier;
			{
				std::lock_guard<std::mutex> ep_lock(req->endpoint->lock);

				// we get an identifier for this request
				identifier = req->endpoint->rdv.add_waitingdata_request(req);


				// find a remote recv slot
				req->rem_rready_message_pos = req->endpoint->rdv.rready_reserve_remote_slot();
				TMPI_ASSERT_FATAL(req->rem_rready_message_pos != -1, "too few rready slots!",
						__FILE__, __LINE__);
			}

			// fill the rready message
			rready_message->send_request = message->send_request;
			rready_message->addr = reinterpret_cast<uint64_t>(req->buf);
			if (req->buflen > 0) { // this can only be set when we have registered the memory
				rready_message->rkey = req->mr->get_mr()->rkey;
			}
			rready_message->recvreq_identifier = identifier;

			ibv_sge sr_sg_lst = {
				reinterpret_cast<uint64_t>(rready_message),         // addr
				rdvRReadyMessageSendSize,                           // length
				VerbsProvider::Rdv::rready_get_send_slots_lkey()    // lkey
			};

			// prepare send request
			ibv_send_wr sr_wr;
			sr_wr.wr_id = reinterpret_cast<uint64_t>(req);
			sr_wr.opcode = IBV_WR_RDMA_WRITE; // we write directly into rdma buffer
			sr_wr.send_flags = IBV_SEND_INLINE | IBV_SEND_SIGNALED;

			sr_wr.sg_list = &sr_sg_lst;
			sr_wr.num_sge = 1; // only one message
			sr_wr.next = nullptr;

			const PeerInfo& pi = req->endpoint->rdv.rready_get_recv_slots_peer_info();
			sr_wr.wr.rdma.rkey = pi.rkey;
			sr_wr.wr.rdma.remote_addr = reinterpret_cast<uint64_t>(pi.addr)
					+ req->rem_rready_message_pos*sizeof(RdvRReadyMessage);

			// set new state before posting
			req->req_common_state = rdv_recv_request_common_states::SENDING_RREADY;
			req->req_rready_state = 0;

			// now we can submit the message to the channel
			ibv_send_wr* bad_wr;
			res = ibv_post_send(req->endpoint->get_qp(), &sr_wr, &bad_wr);
			TMPI_ASSERT_FATAL(res == 0, "ibv_post_send() returned " << res, __FILE__, __LINE__);
			TMPI_LOG_LEVEL(3, "ibv_post_send", "with imm_data = " << sr_wr.imm_data);

			// post a dummy receive because we are waiting for the data
			req->endpoint->post_dummy_recv();

			return 0;
		}

		/////////////////////////////
		// primary state changes
		// primary state changes when the rready has been sent
		inline int rready_sendingrready_to_done(VerbsRdvRecvRequest* const req) {
			// no lock required

			VerbsProvider::Rdv::rready_free_send_slot(req->send_rready_slot_pos);

			// two events are needed for a request to get from sendingrready to done. therefore,
			// we are only completed when fetch_add returns 1 and state is increased to 2
			if (req->req_rready_state.fetch_add(1) == 1) {
				req->req_common_state = rdv_recv_request_common_states::DONE;

				using namespace tinympi::impl;
				baseRequest::request_completed(TO_BaseRequest(req));
				return 1;
			} else {
				return 0;
			}
		}

		inline int common_sendingrready_to_done(VerbsRdvRecvRequest* const req) {

			// if we have received the data, we can free the remote slot which has been consumed now
			// and we can deregister our memory region, because we have received the data
			req->endpoint->rdv.rready_free_remote_slot(req->rem_rready_message_pos);

			req->mr.reset();

			// two events are needed for a request to get from sendingrready to done. therefore,
			// we are only completed when fetch_add returns 1 and state is increased to 2
			if (req->req_rready_state.fetch_add(1) == 1) {
				req->req_common_state = rdv_recv_request_common_states::DONE;

				using namespace tinympi::impl;
				baseRequest::request_completed(TO_BaseRequest(req));
				return 1;
			} else {
				return 0;
			}
		}

	} // END NAMESPACE impl

	/////////////////////////////
	// transparent function

	// there are two main flows:
	//   post to waiting queue --->  retrieve sready
	//     ^ in base request

	//   retrieve sready       -->   get request

	// this function is called when we already have a waiting request, and now an sready has been
	// found in the sready slots
	inline int retrieved_sready_for_waiting_request(VerbsRdvRecvRequest* const req,
			const RdvSReadyMessage* const message, VerbsEndpoint* const endpoint) {
		// no lock required
		using namespace impl;
		using namespace tinympi::impl;

		// we need to cast the base request to a rdv recv request
		req->provider_type = provider_types::node2node;
		req->verbs_request_type = verbs_request_types::VERBS_RDVRECV;
		req->req_common_state = rdv_recv_request_common_states::WAITING_SREADY;
		req->endpoint = endpoint;
		req->rid = message->rid;
		req->peer_rank = message->src_id;
		memset(&req->mr, 0, sizeof(req->mr));

		return waitingsready_to_sendingrready(req, message);
	}
	// this function is called when we received data (written with immediate). Dispatch to the right
	// method
	inline int retrieved_data(VerbsRdvRecvRequest* const req) {
		// no lock required

		using namespace impl;
		return common_sendingrready_to_done(req);
	}

	// this function is called whenever we poll an rready sent event from the completion queue
	inline int handle_poll(VerbsRdvRecvRequest* const req, const ibv_wc&) {
		// no lock required

		using namespace impl;
		return rready_sendingrready_to_done(req);
	}

	// when we retrieve an unexpected rdv message, we generate an rdv recv request from this
	// message, which is then added to the unexpected queue
	inline VerbsRdvRecvRequest* retrieved_unexpected_rdv(int message_pos,
			const RdvSReadyMessage* const message, VerbsEndpoint* const endpoint) {
		// no lock required
		using namespace impl;

		// try to get a request, otherwise allocate a new one
		VerbsRdvRecvRequest* req = TO_VerbsRdvRecvRequest(baseRequest::malloc());
		init(req);

		INIT_BASE_REQUEST(
			req,
			message->src_id,
			message->context_id,
			message->tag,
			nullptr,
			message_pos, // we initialize the size of the request with the message position
			nullptr \
		)

		req->endpoint = endpoint;
		req->rid = message->rid;

		req->req_common_state = rdv_recv_request_common_states::UNEXPECTED_REQUEST;
		return req;
	}
	// a request has been posted for an eager message. We already know that req is rdv!
	inline int posted_request_for_unexpected_rdv(VerbsRdvRecvRequest* const req, void* const buf,
			buflen_t buflen) {
		// no lock required
		using namespace impl;

		// we restore our message from the buffer
		int message_pos = req->buflen;
		const RdvSReadyMessage* message = req->endpoint->rdv.sready_get_recv_slot(message_pos);

		// set the request to the size given by the local recv (overflow test)
		req->buflen = buflen;
		req->buf = buf;
		req->waiting_task = tasks::get_current();
		// tag, context_id, peer already match

		// we are no longer in the UNEXPECTED_REQUEST state
		req->req_common_state = rdv_recv_request_common_states::WAITING_SREADY;

		// then just retrieved the sready as usual
		waitingsready_to_sendingrready(req, message);

		// the recv slot is available again (no lock needed)
		req->endpoint->rdv.sready_free_recv_slot(message_pos);

		return TMPI_SUCCESS;
	}


} // END NAMESPACE verbsRdvRecvRequest
} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif //_VERBS_RDVRECVREQUEST_H_
