// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _THREADS_INIT_H_635106f3226c46d3d114158785752466
#define _THREADS_INIT_H_635106f3226c46d3d114158785752466

#include <atomic>
#include "utility/error.h"
#include <cstring>
#include <vector>
#include <thread>
#include <iostream>

#include <malloc.h>
#include <sched.h>
#include <sys/types.h>
#include <unistd.h>

#include "global.h"

#include "stack.h"
#include "thread_env.h"

namespace tinympi {
namespace threads {
namespace impl {

	// this routine is called by pthreads, when we start a new thread. it is given a void pointer,
	// which points to the thread context
	inline void* thread_startup(void* th) {
		// recover thread context
		thread_context* tctx = (thread_context*) th;
		self = tctx->my_thread;
		env::thread_id = self->rank;

		// allocate the actual stack we will use for running the program
		self->stack = memalign(64, stack_size);
		self->stack_top = reinterpret_cast<void*>(reinterpret_cast<uint8_t*>(self->stack)
				+ stack_size);

        // we make the assumption that any references on the stack are aligned to 8 byte boundaries.
		uint64_t* state_copy = reinterpret_cast<uint64_t*>(tctx->state_copy);

		// state_copy stores first the size of the stack it has copied, then the old stack pointer,
		// and finally the copy itself
		uint64_t stack_src_size = state_copy[0]; // in bytes
		uint64_t old_stack_ptr = state_copy[1]; // store as uint64_t because not needed as pointer
		uint64_t* stack_src_start = &state_copy[2];

		// we need to copy the stack to the end of our new stack
		uint64_t* stack_dest_start = reinterpret_cast<uint64_t*>(
				reinterpret_cast<uint64_t>(self->stack) + stack_size - stack_src_size);

		// then copy the stack
		memcpy(stack_dest_start, stack_src_start, stack_src_size);

		// we need to rewrite the stack. All references to the old stack need to point to the new
		// stack. Without this step we get into trouble.
		uint64_t stack_start_diff = reinterpret_cast<uint64_t>(stack_dest_start) - old_stack_ptr;
		for (unsigned i = 0; i < stack_src_size / 8; ++i) {
			// if we between the old stack bottom and the old stack top, we rewrite the reference.
			if (stack_dest_start[i] >= old_stack_ptr
					&& stack_dest_start[i] < old_stack_ptr + stack_src_size) {
				stack_dest_start[i] = stack_dest_start[i] + stack_start_diff;
			}
		}

		// now we will restore the copy of the stack as our own stack. we will return to init,
		// with the thread context as return value
		restore_stack(stack_dest_start, &self->original_stackptr, tctx);

		////////////////////////////////////////////////////////////
		////////////////////////////////////////////////////////////

		// we will get to this point when we terminate the thread. the copied stack and this thread
		// are no longer necessary

		// the stack is no more needed
		free(self->stack);

		// a return value is not available.
		pthread_exit(nullptr);
		return nullptr;
	}

}

inline void terminate() {
	using namespace impl;

	if (self == &threads[0]) {
		// if we are the main thread, we wait for all other threads
		for (unsigned i = 1; i < threads.size(); ++i) {
			pthread_join(threads[i].thread, nullptr);
		}

	} else {
		// otherwise, we recover the original pthreads stack and return
		recover_threadstack(self->original_stackptr);

		// and never come back ...
		throw std::logic_error("came back from terminate!");
	}
}

// this function returns true, if we are the main thread, and false otherwise
inline bool startup() {
	using namespace impl;

	// initialize global variables
	int res;
	thread_id_t num_threads = env::num_threads;
	threads.resize(num_threads);
	barrier_.init();

	// get the hostname for debugging purposes
	char hostname[64];
	res = gethostname(hostname, sizeof(hostname));
	TMPI_ASSERT_FATAL_EFF(res == 0, "Could not get hostname", __FILE__, __LINE__);
	std::cout << "hostname: " << hostname << std::endl;

	/////////////////////////////////////////////
	// detect environment

	// get the cpu set of the current process
	cpu_set_t cpuset_recv;
	res = sched_getaffinity(getppid(), sizeof(cpu_set_t), &cpuset_recv);
	// count how many cores we can bind to
	unsigned num_cores = env::num_cores;

	// at most we want one thread per cores, otherwise abort
	TMPI_LOG_LEVEL(3, "number of threads: ", (int) num_threads);
	TMPI_LOG_LEVEL(3, "number of cores: ", num_cores);
	TMPI_ASSERT_FATAL_EFF(num_cores >= num_threads, "Cannot have more threads than cores", __FILE__,
			__LINE__);

	// the main proc will bind to the first core available
	cpu_set_t cpuset_new;
	unsigned c = 0;
	while (c < num_cores)
		if (CPU_ISSET(c, &cpuset_recv))
			break;
		else
			c += 1;
	unsigned mainproc_core = c;

	// we need to find the stack of the current process
	pthread_attr_t attr;
	pthread_t p = pthread_self();
	res = pthread_getattr_np(p, &attr);
	TMPI_ASSERT_FATAL_EFF(res == 0, "pthread_getattr_np() failed", __FILE__, __LINE__);

	void* base_stack_addr;
	size_t base_stack_size;
	res = pthread_attr_getstack(&attr, &base_stack_addr, &base_stack_size);
	TMPI_ASSERT_FATAL_EFF(res == 0, "pthread_attr_getstack() failed", __FILE__, __LINE__);

	// calculate the top of the stack (respectively above the top)
	uint64_t* base_stack_top = reinterpret_cast<uint64_t*>(
			reinterpret_cast<uint64_t>(base_stack_addr) + base_stack_size);

	// the barrier count delays the main thread until all other threads have copied the state
	std::atomic_uint* barrier_cnt = new std::atomic_uint(0);

	// then store the current stack. several threads will return from this function. The return value
	// is the context, we will give thread_startup below
	void* state = malloc(stack_size);
	thread_context* restored_ctx = copy_stack(state, base_stack_top);
	if (restored_ctx != nullptr) {

		// the context is no longer needed
		free(restored_ctx);

		// notify barrier
		(*barrier_cnt)++;

		// when we are a new thread
		return false;
	}

	// otherwise continue as main thread

	// then start one thread on each remaining core
	res = pthread_attr_init(&attr);
	TMPI_ASSERT_FATAL_EFF(res == 0, "pthread_attr_init() failed", __FILE__, __LINE__);

	for (thread_id_t th = 1; th < num_threads; ++th) {

		// create one thread context for each thread and set its values
		thread_context* tctx = reinterpret_cast<thread_context*>(malloc(sizeof(thread_context)));
		tctx->my_thread = &threads[th];
		tctx->state_copy = state;
		tctx->my_thread->rank = th;

		// create thread
		res = pthread_create(&threads[th].thread, &attr, &impl::thread_startup, (void*) tctx);
		TMPI_ASSERT_FATAL_EFF(res == 0, "pthread_create() failed", __FILE__, __LINE__);

		// get a free core
		c += 1;
		while (c < num_cores)
			if (CPU_ISSET(c, &cpuset_recv))
				break;
			else
				c += 1;
		TMPI_DASSERT_FATAL(c < num_cores, "# threads should be at least # cores", __FILE__, __LINE__);

		// and bind it to one core
		CPU_ZERO(&cpuset_new);
		CPU_SET(c, &cpuset_new);
		res = pthread_setaffinity_np(threads[th].thread, sizeof(cpu_set_t), &cpuset_new);
		TMPI_ASSERT_FATAL_EFF(res == 0, "pthread_setaffinity_np() failed", __FILE__, __LINE__);
	}

	// now we can also bind this core
	CPU_ZERO(&cpuset_new);
	CPU_SET(mainproc_core, &cpuset_new);
	res = sched_setaffinity(getpid(), sizeof(cpu_set_t), &cpuset_new);
	TMPI_ASSERT_FATAL_EFF(res == 0, "sched_setaffinity() failed", __FILE__, __LINE__);

	res = pthread_attr_destroy(&attr);
	TMPI_ASSERT_FATAL_EFF(res == 0, "pthread_attr_destroy() failed", __FILE__, __LINE__);

	// setup thread 0
	threads[0].rank = 0;
	threads[0].stack_top = base_stack_top;
	self = &threads[0];

	(*barrier_cnt)++;
	while (*barrier_cnt != num_threads) {}
	delete barrier_cnt;

	// when all threads have started, we can free the state
	free(state);

	return true;

}

} // END NAMESPACE threads
} // END NAMESPACE tinympi

#endif // _THREADS_INIT_H_

