#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "malloc_overwrite.h"

#include <thread>
#include <mutex>
#include <atomic>
#include <dlfcn.h>

static real_malloc_t libc_malloc = nullptr;
static real_memalign_t libc_memalign = nullptr;

static hook_malloc_t hook_malloc = nullptr;
static std::recursive_mutex global_lock;
static std::atomic<std::thread::id> global_lock_owner;

static void memalign_init() {
	libc_memalign = (real_memalign_t) dlsym(RTLD_NEXT, "memalign");
	if (libc_memalign == nullptr) {
		// here we are allowed to print, because malloc does not exist, so we will fail anyway
		fprintf(stderr, "Error in dlsym: %s\n", dlerror());
		return;
	}
}
static void malloc_init() {
	libc_malloc = (real_malloc_t) dlsym(RTLD_NEXT, "malloc");
	if (libc_malloc == nullptr) {
		// here we are allowed to print, because malloc does not exist, so we will fail anyway
		fprintf(stderr, "Error in dlsym: %s\n", dlerror());
		return;
	}
}

extern "C"
void set_malloc_hook_(hook_malloc_t hook_function) {
	hook_malloc = hook_function;
}

extern "C"
void* memalign(size_t alignment, size_t size) {
	if (libc_memalign == nullptr)
		memalign_init();

	void* ptr = libc_memalign(alignment, size);
	if (hook_malloc != nullptr && !(global_lock_owner == std::this_thread::get_id()))
		hook_malloc(ptr, size);
	return ptr;
}

extern "C"
void* real_memalign_(size_t alignment, size_t size) {
	if (libc_memalign == nullptr)
		memalign_init();

	std::lock_guard<std::recursive_mutex> lk(global_lock);
	global_lock_owner = std::this_thread::get_id();
	void* res = libc_memalign(alignment, size);
	global_lock_owner = std::thread::id();

	return res;
}

extern "C"
void* malloc(size_t sz) noexcept {
	if (libc_malloc == nullptr)
		malloc_init();

	void* ptr = libc_malloc(sz);
	if (hook_malloc != nullptr && !(global_lock_owner == std::this_thread::get_id()))
		hook_malloc(ptr, sz);
	return ptr;
}

extern "C"
void* real_malloc_(size_t sz) {
	if (libc_malloc == nullptr)
		malloc_init();

	std::lock_guard<std::recursive_mutex> lk(global_lock);
	global_lock_owner = std::this_thread::get_id();
	void* res = libc_malloc(sz);
	global_lock_owner = std::thread::id();
	return res;
}
